package sam.fol ; 

import sam.lang.* ; 
import java.util.* ; 
import java.awt.* ;  // for Point

// Typed quantifiers
// Consistency rules only for negated atoms in consequent of geolog rule
// No more automatic closures so can remove the parameter
// auto dom

/**
  * A WFF is a pseudo-tree of logical forms.
  * Fore-and-Aft translation. 
  * @version 22 May 2011  Coherent Fore-and-aft Translations 
  * @author &copy; John Fisher 
  */
public class WFF { 

   // strings designating fore/aft translation
   public static String FORE = "<" ;    
   public static String AFT  = ">" ;  

   //////////////////////////
   // types of wff
   //////////////////////////
   public static int FUNCTOR = 0 ; //    w
   public static int AND = 1 ;     //    w    w ...
   public static int OR = 2 ;      //    w    w ...
   public static int NOT = 3 ;     //    w
   public static int IF = 4 ;      //    w    w
   public static int IFF = 5 ;     //    w    w
   public static int ALL = 6 ;     //   vars  w
   public static int EXISTS = 7 ;  //   vars  w

   //////////////////////////////////////
   //     op      operands
   //  FUNCTOR: [functor], <br>
   //  AND: [wff,wff,..], <br>
   //  OR: [wff,wff,..], <br>
   //  NOT: [wff] , <br>
   //  IF: [wff,wff] , <br>
   //  IFF: [wff,wff] , <br>
   //  ALL: [vars,wff] , <br>
   //  EXISTS: [vars,wff] , <br>  
   /////////////////////////////////////// 
   public int op ; 
   public Vector<Object> operands ;  // multi class

   // The universal domain name
   private static String DOM = "dom" ;  


   /**
     *  (partially) Construct WFF knowing 
     *  logical type (AND/OR/NOT/...).
     *  Then add the operands.  
     */
   public WFF(int t) { 
      this.op = t ;
      this.operands = new Vector<Object>() ;
   }

   /**
     *  toString()
     */
   public String toString() { 
      StringBuffer s = new StringBuffer() ; 
      if (this.op == 0) { // FUNCTOR  -- needed modification
          s.append(((Functor)this.operands.elementAt(0)).toFOLString()) ;
      }
      else if (this.op == 1) {  //AND
         s.append("(") ; 
         Iterator it = this.operands.iterator() ; 
         s.append(((WFF)it.next()).toString()) ;
         while (it.hasNext()) { 
            s.append(" & ") ;
            s.append(((WFF)it.next()).toString()) ; 
         }
         s.append(")") ; 
      }
      else if (this.op == 2) {  //OR
         s.append("(") ; 
         Iterator it = this.operands.iterator() ; 
         s.append(((WFF)it.next()).toString()) ;
         while (it.hasNext()) { 
            s.append(" | ") ;
            s.append(((WFF)it.next()).toString()) ; 
         }
         s.append(")") ; 
      }
      else if (this.op == 3) {  //NOT
         s.append("~") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
      }
      else if (this.op == 4) {  //IF
         s.append("(") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
         s.append(" => ") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
         s.append(")") ; 
      }
      else if (this.op == 5) {  //IFF
         s.append("(") ; 
         s.append(((WFF)this.operands.elementAt(0)).toString()) ;
         s.append(" <=> ") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
         s.append(")") ; 
      }
      else if (this.op == 6) {  //ALL
         s.append("![") ; 
         Iterator vars = ((Vector)this.operands.elementAt(0)).iterator() ;
         s.append(((Variable)(vars.next())).toString()) ;
         while(vars.hasNext()) { 
            s.append(",") ;
            s.append(((TypedVariable)(vars.next())).toString()) ; 
         }
         s.append("]:") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
      }
      else if (this.op == 7) {  //EXISTS
         s.append("?[") ; 
         Iterator vars = ((Vector)this.operands.elementAt(0)).iterator() ;
         s.append(((TypedVariable)(vars.next())).toString()) ;
         while(vars.hasNext()) { 
            s.append(",") ;
            s.append(((Variable)(vars.next())).toString()) ; 
         }
         s.append("]:") ; 
         s.append(((WFF)this.operands.elementAt(1)).toString()) ;
      }   
      return s.toString() ;
   }

   /**
     *  @return logical negation of this WFF
      */
   public WFF negation() {
      WFF nwff = new WFF(NOT) ; 
      nwff.operands.add(this) ; 
      return nwff ;
   }


   /**
     *  Find the variables in a CNF. 
      */
   public HashMap findVariables() { 
      HashMap<String,Integer> m = new HashMap<String,Integer>() ; 
      if (this.op == FUNCTOR) {  
         Functor f = (Functor)this.operands.elementAt(0) ; 
         f.findVariables(m) ;  
      }
      else if (this.op == NOT) { 
         WFF w = (WFF)this.operands.elementAt(0) ;
         Functor f = (Functor)w.operands.elementAt(0) ; 
         f.findVariables(m) ;  
      }     
      else {  // OR
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext()) { 
            WFF w = (WFF)it.next() ; 
            if (w.op == FUNCTOR) {  
               Functor f = (Functor)w.operands.elementAt(0) ; 
               f.findVariables(m) ;  
            }
            else if (w.op == NOT) { 
               WFF w1 = (WFF)w.operands.elementAt(0) ;
               Functor f = (Functor)w1.operands.elementAt(0) ; 
               f.findVariables(m) ;  
            }     
         }
      } 
      return m ; 
   }

///////////////////////////////////////////////////////////////
///////// CONVERT TO GEOLOG RULES
/////////   coherent rules, domain declarations and closures

   /**
     *  Convert to geolog theory, 
     *  positive orientation.
      */
   public void toGeolog(Theory thy,Theory consistency,TreeSet<String> guards) { 
      // assert reified wff (+)
      Functor tru = new Functor("true",0) ; 
      Conjunction a = new Conjunction() ; 
      a.add(tru) ; 
      Conjunction b = new Conjunction() ; 
      VarList uvars = new VarList() ; 
      b.add(this.preify(AFT,uvars)) ;  // no vars yet
      Disjunction d = new Disjunction() ; 
      d.add(b) ; 
      Rule rule = new Rule(a,d) ; 
      thy.add(rule) ; 
      this.pAft(thy,consistency,uvars,guards) ;  
System.out.println(guards) ;
   }

   protected void pAft(Theory thy,Theory consistency,VarList uvars,TreeSet<String> guards) { 
      if (this.op == FUNCTOR) {
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            b.add(((WFF)it.next()).preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(b) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pAft(thy,consistency,(VarList)uvars.clone(),guards); //N.B. copy
      }
      else if (this.op == OR) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ; 
         Iterator it = this.operands.iterator() ; 
         Disjunction d = new Disjunction() ; 
         while (it.hasNext())  {
            Conjunction c = new Conjunction() ;
            c.add(((WFF)it.next()).preify(AFT,uvars)) ; 
            d.add(c) ;
         }
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pAft(thy,consistency,(VarList)uvars.clone(),guards); 
      }
      else if (this.op == NOT) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ; 
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ;
         c.add(w.nreify(AFT,uvars)) ; 
         d.add(c) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.nAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == IF) {   
        // w1 => w2
        WFF w1 = (WFF)this.operands.elementAt(0) ;
        WFF w2 = (WFF)this.operands.elementAt(1) ;
        Theory tmp_thy = new Theory(),
               tmp_consistency = (Theory)(consistency.clone()) ;
        boolean fore_try = true ; 
        // attempt fore translation of w1
        try { // try to fore-translate antecedent
           w1.pFore(tmp_thy,tmp_consistency,(VarList)uvars.clone(),guards) ; 
        } 
        catch(ForeException fex) {  fore_try = false ;  }
        if (fore_try) { // Fore translation  was ok
            Conjunction ant = new Conjunction() ; 
            ant.add(this.preify(AFT,uvars)) ; 
            ant.add(w1.preify(FORE,uvars)) ;  
            Disjunction d = new Disjunction() ; 
            Conjunction c = new Conjunction() ; 
            c.add(w2.preify(AFT,uvars)) ; 
            d.add(c) ; 
            Rule rule = new Rule(ant,d) ; 
            thy.add(rule) ; 
            // update from fore-translation
            thy.addAll(tmp_thy) ;   
            consistency.clear() ; // and transfer 
            consistency.addAll(tmp_consistency) ; 
            // w2 last, below
        }
        else {   // Fore translation failed
            Conjunction ant = new Conjunction() ; 
            ant.add(this.preify(AFT,uvars)) ; 
            Disjunction d = new Disjunction() ; 
            Conjunction c1 = new Conjunction() ; 
            Conjunction c2 = new Conjunction() ; 
            c1.add(w1.nreify(AFT,uvars)) ; 
            c2.add(w2.preify(AFT,uvars)) ; 
            d.add(c1) ; d.add(c2) ; 
            Rule rule = new Rule(ant,d) ; 
            thy.add(rule) ; 
            w1.nAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
         }
         // and finally
         w2.pAft(thy,consistency,(VarList)uvars.clone(),guards) ;  
      }
      else if (this.op == IFF) {  
         // >(a<=>b)=> >(a=>b), >(b=>a)
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         WFF w3 = new WFF(IF) ; 
         w3.operands.add(w1) ; w3.operands.add(w2) ;
         WFF w4 = new WFF(IF) ;
         w4.operands.add(w2) ; w4.operands.add(w1) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(w3.preify(AFT,uvars)) ;  
         c.add(w4.preify(AFT,uvars)) ;  
         d.add(c) ; 
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w3.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
         w4.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == ALL) {  
         // '+{![X:a,Y:b..]: W}'(..uvars..), a(X), b(Y) .. => '+{W}'(..uvars..,X,Y..).
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ;
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < ((Vector)this.operands.elementAt(0)).size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1) ;
            e.args[0]= v.clone() ;
            ant.add(e) ; 
         }
         // NOW include the new universal variables in uvars
         Iterator newvars = ((Vector)this.operands.elementAt(0)).iterator() ; 
         while (newvars.hasNext()) uvars.add(((Variable)newvars.next()).clone()) ; 
         Conjunction b = new Conjunction() ; 
         WFF w = (WFF)(this.operands.elementAt(1)) ;
         //Functor f = new Functor(w.preify(AFT,uvars),0) ; 
         b.add(w.preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == EXISTS) {  
         // '+{?[X:a,Y:b..]: W}'(..uvars..) => a(X), b(Y).., '+{W}'(..uvars..,X,Y..).
         Conjunction ant = new Conjunction() ; 
         ant.add(this.preify(AFT,uvars)) ;
         Conjunction b = new Conjunction() ; 
         Vector tvars = (Vector)this.operands.elementAt(0) ;
         for(int i=0 ; i < ((Vector)this.operands.elementAt(0)).size() ; i++) { 
            TypedVariable v = (TypedVariable)tvars.elementAt(i) ; 
            guards.add(v.typename) ; 
            Functor e = new Functor(v.typename,1) ;
            e.args[0]= v.clone() ; 
            b.add(e) ; 
         }
         // NOW include the new variables in uvars
         Iterator newvars = ((Vector)this.operands.elementAt(0)).iterator() ; 
         while (newvars.hasNext()) uvars.add(((Variable)newvars.next()).clone()) ; 
         WFF w = (WFF)(this.operands.elementAt(1)) ; 
         b.add(w.preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;
         thy.add(new Rule(ant,d)) ; 
         w.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
   } // end pAft

   protected void nAft(Theory thy,Theory consistency,VarList uvars,TreeSet<String> guards) { 
      if (this.op == FUNCTOR) {
         // The negation of a predicate in the aft direction
         consistency( (Functor)this.operands.elementAt(0),consistency) ; 
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         Disjunction d = new Disjunction() ; 
         while (it.hasNext())  {
            Conjunction c = new Conjunction() ;
            c.add(((WFF)it.next()).nreify(AFT,uvars)) ; 
            d.add(c) ;
         }
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nAft(thy,consistency,(VarList)uvars.clone(),guards); 
      }
      else if (this.op == OR) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ; 
         Conjunction b = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            b.add(((WFF)it.next()).nreify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(b) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nAft(thy,consistency,(VarList)uvars.clone(),guards); 
      }
      else if (this.op == NOT) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ; 
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ;
         c.add(w.preify(AFT,uvars)) ; 
         d.add(c) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == IF) {   
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         c.add(w1.preify(AFT,uvars)) ; c.add(w2.nreify(AFT,uvars)) ; 
         d.add(c) ; 
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w1.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
         w2.nAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == IFF) {   
         // >-(a<=>b) => >-(a=>b) | >-(b=>a)
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         WFF w3 = new WFF(IF) ;
         w3.operands.add(w1) ; w3.operands.add(w2) ; 
         WFF w4 = new WFF(IF) ;
         w4.operands.add(w2) ; w4.operands.add(w1) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction c1 = new Conjunction() ; 
         Conjunction c2 = new Conjunction() ; 
         c1.add(w3.nreify(AFT,uvars)) ;  
         c2.add(w4.nreify(AFT,uvars)) ;  
         d.add(c1) ; d.add(c2) ; 
         thy.add( new Rule(ant,d)) ; 
         w3.nAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
         w4.nAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == ALL) {  
         //    ___this____
         // '-{![X,Y..]: W}'(..uvars..) => '+{?[X,Y..]: ~W}'(..uvars). 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ;
         WFF neg = new WFF(NOT) ; 
         neg.operands.add(this.operands.elementAt(1)) ;
         WFF w = new WFF(EXISTS) ; 
         w.operands.add(this.operands.elementAt(0)) ;
         w.operands.add(neg) ; 
         Conjunction b = new Conjunction() ; 
         b.add(w.preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
      else if (this.op == EXISTS) {  
         // '-{?[X,Y..]: W}'(..uvars..) => '+{![X,Y..]: ~W}'(..uvars). 
         Conjunction ant = new Conjunction() ; 
         ant.add(this.nreify(AFT,uvars)) ;
         WFF neg = new WFF(NOT) ; 
         neg.operands.add(this.operands.elementAt(1)) ;
         WFF w = new WFF(ALL) ; 
         w.operands.add(this.operands.elementAt(0)) ;
         w.operands.add(neg) ; 
         Conjunction b = new Conjunction() ; 
         b.add(w.preify(AFT,uvars)) ; 
         Disjunction d = new Disjunction() ;
         d.add(b) ;  
         thy.add(new Rule(ant,d)) ; 
         w.pAft(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
   }

   /**
     *  positive Fore translations
     *  @throws ForeException
      */
   protected void pFore(Theory thy,Theory consistency,VarList uvars,TreeSet<String> guards) { 
      if (this.op == FUNCTOR) {  // capture free variables in antecedent
         // p(X), dom(Y), dom(Z) => <{p(X)} (X,Y,Z)  
         // find variables in uvars that are not in the functor
         Functor f = (Functor)this.operands.elementAt(0) ; 
         HashMap<String,Integer> m = new HashMap<String,Integer>() ; 
         f.findVariables(m) ;  
         Set vars = m.keySet() ; 
         VarList freevars = new VarList() ;  
         Iterator it = uvars.iterator() ; 
         while (it.hasNext()) {
            Variable v = (Variable)it.next() ;
            if (!vars.contains(v)) freevars.add(v.clone()) ; 
         } 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(this.preify(FORE,uvars)) ; 
         d.add(c) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(f) ; 
         Iterator q = freevars.iterator() ;
         while (q.hasNext()) {  
            Functor dom = new Functor(DOM,1) ;
            dom.args[0] = ((Variable)q.next()).clone() ; 
            ant.add(dom) ; 
         }         
         thy.add(new Rule(ant,d)) ; 
      }
      else if (this.op == AND) {   
         Conjunction ant = new Conjunction() ; 
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())   
            ant.add(((WFF)it.next()).preify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction cons = new Conjunction() ; 
         cons.add(this.preify(FORE,uvars)) ; 
         d.add(cons) ;  
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pFore(thy,consistency,(VarList)uvars.clone(),guards); //N.B. copy
      }
      else if (this.op == OR) {   
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,uvars)) ; 
         d.add(conj) ;
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext())  {
            Conjunction ant = new Conjunction() ;
            ant.add(((WFF)it.next()).preify(FORE,uvars)) ; 
         thy.add(new Rule(ant,d)) ; 
         }
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).pFore(thy,consistency,(VarList)uvars.clone(),guards); 
      }
      else if (this.op == NOT) {   
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Conjunction ant = new Conjunction() ;
         ant.add(w.nreify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,uvars)) ; 
         d.add(conj) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }   
      else if (this.op == IF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         // w1 => w2
         // <-w1 => <(w1 => w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(conj) ; 
         // <w2 => <(w1 => w2)
         Conjunction ant1 = new Conjunction() ; 
         ant1.add(w1.nreify(FORE,uvars)) ;
         thy.add(new Rule(ant1,d)) ;
         Conjunction ant2 = new Conjunction() ; 
         ant2.add(w2.preify(FORE,uvars)) ;
         thy.add(new Rule(ant2,d)) ;
         w1.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
         w2.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
      else if (this.op == IFF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         //  <(w1=>w2), <(w2=>w1) => <(w1<=>w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.preify(FORE,uvars)) ;
         Disjunction d = new Disjunction() ;
         d.add(conj) ; 
         Conjunction ant = new Conjunction() ; 
         WFF A = new WFF(IF) ; 
         A.operands.add(w1) ;          
         A.operands.add(w2) ;          
         WFF B = new WFF(IF) ; 
         B.operands.add(w2) ;          
         B.operands.add(w1) ;          
         ant.add(A.preify(FORE,uvars)) ; 
         ant.add(B.preify(FORE,uvars)) ; 
         thy.add(new Rule(ant,d)) ; 
         A.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
         B.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
      else if (this.op == ALL) {  
         // EXCEPTIONAL CASE
         throw new ForeException(this) ; 
         /////////////////////////////////
      } 
      else if (this.op == EXISTS) {  
         // <w => <(Ex)w
         WFF w = (WFF)this.operands.elementAt(1) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(w.preify(FORE,uvars)) ;
         Conjunction b = new Conjunction() ; 
         b.add(this.preify(FORE,uvars)) ;
         Disjunction d = new Disjunction() ;
         d.add(b) ;
         thy.add(new Rule(ant,d)) ; 
         w.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }
   } // end pFore

   /**
     *  negative Fore translations
     *  @throws ForeException
      */
   protected void nFore(Theory thy,Theory consistency,VarList uvars,TreeSet<String> guards) { 
      if (this.op == FUNCTOR) {  // capture free variables in antecedent
         // not_p(X), dom(Y), dom(Z) => <-{p(X)} (X,Y,Z)  
         // find variables in uvars that are not in the functor
         Functor f = (Functor)this.operands.elementAt(0) ; 
         HashMap<String,Integer> m = new HashMap<String,Integer>() ; 
         f.findVariables(m) ;  
         Set vars = m.keySet() ; 
         VarList freevars = new VarList() ;  
         Iterator it = uvars.iterator() ; 
         while (it.hasNext()) {
            Variable v = (Variable)it.next() ;
            if (!vars.contains(v)) freevars.add(v.clone()) ; 
         } 
         Disjunction d = new Disjunction() ; 
         Conjunction c = new Conjunction() ; 
         c.add(this.nreify(FORE,uvars)) ; 
         d.add(c) ; 
         Conjunction ant = new Conjunction() ; 
         ant.add(explicitly_negate(f)) ;   ///// N.B. /////    not_p => ...
         Iterator q = freevars.iterator() ;
         while (q.hasNext()) {   
            Functor dom = new Functor(DOM,1) ;
            dom.args[0] = ((Variable)q.next()).clone() ; 
            ant.add(dom) ; 
         }         
         thy.add(new Rule(ant,d)) ; 
      }   
      else if (this.op == AND) {   
         Conjunction cons = new Conjunction() ; 
         cons.add(this.nreify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(cons) ;  
         Iterator it = this.operands.iterator() ; 
         while (it.hasNext()) {
            Conjunction ant = new Conjunction() ; 
            ant.add(((WFF)it.next()).nreify(FORE,uvars)) ;
            thy.add(new Rule(ant,d)) ; 
         }
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nFore(thy,consistency,(VarList)uvars.clone(),guards); //N.B. copy
      }
      else if (this.op == OR) {   
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,uvars)) ; 
         d.add(conj) ;
         Iterator it = this.operands.iterator() ; 
         Conjunction ant = new Conjunction() ;
         while (it.hasNext())  {
            ant.add(((WFF)it.next()).nreify(FORE,uvars)) ; 
         }
         thy.add(new Rule(ant,d)) ; 
         it = this.operands.iterator() ; 
         while (it.hasNext())   
            ( (WFF)it.next() ).nFore(thy,consistency,(VarList)uvars.clone(),guards); 
      }   
      else if (this.op == NOT) {   
         WFF w = (WFF)this.operands.elementAt(0) ; 
         Conjunction ant = new Conjunction() ;
         ant.add(w.preify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,uvars)) ; 
         d.add(conj) ;
         Rule rule = new Rule(ant,d) ; 
         thy.add(rule) ; 
         w.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      }  
      else if (this.op == IF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         // <w1,<-w2 => <-(w1 => w2)
         Conjunction ant = new Conjunction() ; 
         ant.add(w1.preify(FORE,uvars)) ;
         ant.add(w2.nreify(FORE,uvars)) ;
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(conj) ; 
         thy.add(new Rule(ant,d)) ;
         w1.pFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
         w2.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
      else if (this.op == IFF) {   
         WFF w1 = (WFF)this.operands.elementAt(0) ;
         WFF w2 = (WFF)this.operands.elementAt(1) ;
         //  <-(w1=>w2) => <-(w1<=>w2)
         //  <-(w2=>w1) => <-(w1<=>w2)
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,uvars)) ;
         Disjunction d = new Disjunction() ;
         d.add(conj) ; 
         Conjunction antA = new Conjunction() ; 
         WFF A = new WFF(IF) ; 
         A.operands.add(w1) ;          
         A.operands.add(w2) ;          
         antA.add(A.nreify(FORE,uvars)) ;
         thy.add(new Rule(antA,d)) ; 
         Conjunction antB = new Conjunction() ; 
         WFF B = new WFF(IF) ; 
         B.operands.add(w2) ;          
         B.operands.add(w1) ;          
         antB.add(B.nreify(FORE,uvars)) ; 
         thy.add(new Rule(antB,d)) ; 
         A.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
         B.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
      else if (this.op == ALL) {  
         // <-{w} => <-{![X]:w
         Conjunction ant = new Conjunction() ; 
         WFF w = (WFF)this.operands.elementAt(1) ;
         ant.add(w.nreify(FORE,uvars)) ; 
         Conjunction conj = new Conjunction() ; 
         conj.add(this.nreify(FORE,uvars)) ; 
         Disjunction d = new Disjunction() ; 
         d.add(conj) ; 
         thy.add(new Rule(ant,d)) ; 
         w.nFore(thy,consistency,(VarList)uvars.clone(),guards) ; 
      } 
      else if (this.op == EXISTS) {  
         // EXCEPTIONAL CASE
         throw new ForeException(this) ; 
         ///////////////////////////////
      }
   } // end nFore

// <<171 (\u00AB) >>187(\u00BB) log.not(\u00AC)
   protected Functor preify(String d, VarList uvars) { // d is FORE or AFT
      if (this.op == FUNCTOR && d == AFT) 
         return (Functor)this.operands.elementAt(0) ; 
      // otherwise decorate 
      String s = "\'"+(d==FORE?"\u00AB":"")+"{" + this.toString() + "}\'" ;
      Functor f =  new Functor(s,uvars.size()) ; 
      for(int i=0 ; i < uvars.size() ; i++) f.args[i] = ((Variable)uvars.elementAt(i)).clone() ; 
      return f ; 
   }

   protected Functor nreify(String d, VarList uvars) {  // d is FORE or AFT
      if (this.op == FUNCTOR && d == AFT)  
         return explicitly_negate((Functor)this.operands.elementAt(0)) ; 
      // otherwise decortate
      String s = "\'"+(d==FORE?"\u00AB~":"~")+"{" + this.toString() + "}\'" ;
      Functor f =  new Functor(s,uvars.size()) ; 
      for(int i=0 ; i < uvars.size() ; i++) f.args[i] = ((Variable)uvars.elementAt(i)).clone() ;
      return f ; 
   }

   /**
     * @return EXPLICIT NEGATION not_f(-) of a functor f(-).
     * Explicit negation of '='/2 is '!='/2.
     */
   protected Functor explicitly_negate(Functor f) { 
      // String n = "'-" + f.name + "'" ;  // old way
      String n  ; 
      if (f.name.equals("=") && f.arity==2) n = "!=" ; 
      else n = "not_" + f.name  ;  
      Functor g = new Functor(n,f.arity) ; 
      for (int i=0 ; i<f.arity ; i++) g.args[i] = f.args[i] ;
      return g ;
   }

   /**
     *  Add p(-), 'not_p'(-) => false. ... if not already present.
     */
   protected void consistency(Functor f, Theory thy) { 
      // no consistency rules for true or dom
      if (f.name.equals("true") || f.name.equals(DOM)) return ; 
      // already present?
      for (int i = 0 ; i < thy.size() ; i++) {
         Rule r = (Rule)thy.elementAt(i) ; 
         Functor g = r.antecedent.elementAt(0) ;
         // compare functors (no index yet)
         if (g.name.equals(f.name) && g.args.length == f.args.length) {
            return ;  // already present so do nothing
         }
      } 
      // otherwise add new consistency rule
      Functor nf = explicitly_negate(f) ;
      Conjunction ant = new Conjunction() ; 
      ant.add(generalize(f)) ; ant.add(generalize(nf)) ; ///
      Disjunction d = new Disjunction() ; 
      Conjunction c = new Conjunction() ; 
      c.add(new Functor("false",0)) ; 
      d.add(c) ; 
      thy.add(new Rule(ant,d)) ; 
   }

   /**
     *  Generate p(X0,X1,...)
     */
   protected Functor generalize(Functor f) { 
      Functor g = new Functor(f.name,f.arity) ; 
      for (int i = 0 ; i < f.arity ; i++) 
         g.args[i] = new Variable("X"+String.valueOf(i)) ; 
      return g ; 
   }


///////////////////////////////////////////////////////////////
/////////
///////// Generate geolog Theory
/////////   
///////////////////////////////////////////////////////////////

   /** 
     *  Generate geolog theory. 
     *  Style is "terse" or "verbose".
     *  No explicit check that there is exactly ONE conjecture.
     *  If there are two conjectures c1 and c2, they are both negated, 
     *  as if the single conjecture had been c1 | c2 (OR conjecture assumption).
      */
   public static Theory geolog(Fol fol,String method) { 
      Theory main = new Theory() , 
             //domain = new Theory(),
             consistency = new Theory() ;
             //closure = new Theory() ;  
      TreeSet<String> guards = new TreeSet<String>() ;  
      ///
      Iterator it = fol.theory.iterator() ; 
      while (it.hasNext()) { 
         Theory temp = new Theory() ; 
         Formula f = (Formula)it.next() ; 
         WFF w = f.wff ;
         boolean coherence = false ; 
         if (f.role.equals("conjecture")) { 
           //if (f.mode.equals("cnf")) w = w.eclosure() ; 
           // negate wff and then translate
           w.negation().toGeolog(temp,consistency,guards) ; 
         }
         else {  // "axiom" etc.
           w.toGeolog(temp,consistency,guards) ;   
         } 
         if ((method.equals("terse") && !coherence) || (method.equals("coherent") && !coherence)) { // TERSE
            Point p = WFF.starts(temp) ; 
            // remove first p.x rules
            for (int i = 0 ; i <= p.x-1 ; i++)   
               temp.remove(0) ;
            // save start rule 
            Rule r = temp.elementAt(0) ; 
            // remove start rule
            temp.remove(0) ; 
            // modify start rule
            Conjunction a = new Conjunction() ; 
            if (p.y == 1) a.add(new Functor("true",0)) ; 
            //else a.add(r.antecedent.elementAt(1)) ; // <<< bug 
            else      // <<< fix 
               for(int i=1 ; i < r.antecedent.size() ; i++) 
                  a.add(r.antecedent.elementAt(i)) ;
            // add modified start rule
            temp.add(0,new Rule(a,r.consequent)) ; 
         ///////////////////////////////////////////////////////////////// 
         temp.indexVariables() ; // added 4.21.2010  going to remove rules using new Term.equals
         temp.indexFunctors() ;  // added 4.21.10
         foreFold(temp) ;     // use folding to rewrite Antecedents of rules
         //System.out.println(temp) ;
         domTrim(temp) ;      // remove unneeded doms from antecedents
         aftFold(temp) ;      // use folding to rewrite consequents of rules
         /////////////////////////////////////////////////////////////////
         }
         // now add temp to main
         if (f.role.equals("conjecture")) 
            for(int i=temp.size()-1  ; i >= 0 ; i--) // reversely
               main.add(0,(Rule)temp.elementAt(i)) ; 
         else {  // forwardly 
            Iterator it2 = temp.iterator() ; 
            while (it2.hasNext()) main.add((Rule)it2.next()) ;          
         }
      }  
      for(int j = consistency.size()-1 ; j >=0 ; j--) 
         main.add(0,(Rule)consistency.elementAt(j)) ; 
      /////
      Iterator grds = guards.iterator() ; 
      while (grds.hasNext()) {
         String s = (String)grds.next() ; 
         //  a(x) => dom(X) for quant guard a
         if (!s.equals(DOM)) main.add(subDom(s)) ; 
      }
      /////
      return main ; 
   }


   /**
     *  a(Z) => dom(Z)  for quant guard a
      */
   protected static Rule subDom(String s) {
      Variable z = new Variable("Z") ; 
      Functor a = new Functor(s,1) ;
      a.args[0] = z ;
      Functor dom = new Functor(DOM,1) ;
      dom.args[0] = z ;
      Conjunction ant = new Conjunction() ;
      ant.add(a) ;
      Conjunction cons = new Conjunction() ; 
      cons.add(dom) ; 
      Disjunction d = new Disjunction() ;
      d.add(cons) ; 
      return new Rule(ant,d) ; 
   }

   /**
     *  Fold the theory using rewriting:
     *  L = <C1 => 'quoty'(X), C2 => 'quoty'(X), ...>
     *  is used to subtitute Ci for 'quoty'(X) 
     *  in antecedents (fore) in all the rules.
      */
   protected static void foreFold(Theory temp) {
      while (true) {          
         Theory quoties = findQuotedConsequents(temp) ;  
         if (quoties == null) break ;  // no more rewrites possible
         // use quoties to rewrite remaining rules
         Functor q = quoties.elementAt(0).consequent.elementAt(0).elementAt(0) ; 
         Rule r = null ; int p = -1 ; int rulenum = -1 ;  
         // find rule/position of rule to rewrite 
         for (int j = 0 ; j < temp.size() ; j++) {
            rulenum = j ; 
            r = temp.elementAt(j) ; 
            p = locate(q,r) ;
            if (p > -1) break ; 
         }
         Theory thy = rewrite(r,quoties,p) ; 
         temp.removeElementAt(rulenum) ; 
         temp.addAll(rulenum,thy) ; // replace r
      }
   }

   // rewrite r using quoties, replacing functor at loc
   static Theory rewrite(Rule r,Theory quoties,int p) {
      Theory thy = new Theory() ; 
      for (int i = 0 ; i < quoties.size() ; i++) {
         Conjunction c = (Conjunction) r.antecedent.clone() ; 
         c.removeElementAt(p) ; 
         c.addAll(p,quoties.elementAt(i).antecedent) ;
         thy.add(new Rule(c,r.consequent)) ;
      }
      return thy ; 
   }

   //remove unneeded doms from antecedents of rules
   static void domTrim(Theory thy) { 
      for (int i = 0 ; i < thy.size() ; i++) { 
         Rule r = thy.elementAt(i) ; 
         if (universal(r.antecedent)) {
            thy.removeElementAt(i) ; // one-for-one
            thy.add(i, new Rule(domTrim(r.antecedent),r.consequent)) ; 
         }
      }
   }

   // does c have a dom in it?
   static boolean universal(Conjunction c) { 
      for (int i = 0 ; i < c.size(); i++) 
         if (c.elementAt(i).name.equals(DOM)) return true ; 
      return false ; 
   }

/*
Folding can also produce repeated doms in an antecedent, e.g.,
dom(X), dom(X) => X=X.
Use domRepeat/2 to detect this.
*/


   //remove any dom(X) if X occurs in another predicate 
   static Conjunction domTrim(Conjunction c) { 
      Conjunction trimmed = new Conjunction() ; 
      for(int i = 0 ; i < c.size() ; i++) {
         Functor f = c.elementAt(i) ; 
         if (f.name.equals(DOM)) {
            Variable v = (Variable)f.args[0] ;
            if (!domRepeat(v,trimmed) && !occurs(v,c))  
               trimmed.add(f) ; // necessary universal
         }
         else trimmed.add(f) ;  // not a dom predicate
      }           
      return trimmed ; 
   }

   static boolean domRepeat(Variable v,Conjunction c) {
      for (int i = 0 ; i < c.size() ; i++) { 
         Functor f = c.elementAt(i) ;
         if(f.name.equals(DOM) && 
            ((Variable)f.args[0]).name.equals(v.name)) 
            return true ; 
      } 
      return false ; 
   }

   // variable occurs in a non-dom context in a conjunction ?
   static boolean occurs(Variable x, Conjunction c) {
      for (int i = 0 ; i < c.size() ; i++) { 
         Functor f = c.elementAt(i) ; 
         if (!f.name.equals(DOM)) {
            for (int j = 0 ; j < f.args.length ; j++)          
               if (occurs(x,(Term)f.args[j])) return true ;  
         }
      }
      return false ;
   }

   static boolean occurs(Variable x, Term t) {
      if (t.isVariable()) {
         if (((Variable)t).name.equals(x.name)) 
            return true ; 
         else return false ;
      }
      else { // t is a functor
         Functor f = (Functor)t ;
         for (int i = 0 ; i < f.args.length ; i++) 
            if (occurs(x,(Term)f.args[i])) return true ; 
      }
      return false ;
   }


   // at what position in antecedent of r does q occur? -1 = doesn't
   static int locate(Functor q, Rule r) { 
      Conjunction c = r.antecedent ;
      for (int i = 0 ; i < c.size(); i++) {
         if (c.elementAt(i).index == q.index) // N.B. uses .index
            return i ; 
      }
      return -1 ; // q not found
   }

   static Theory findQuotedConsequents(Theory temp) {
      Theory quoties = new Theory() ;  
      Functor quoty = null ; // will find
      for (int i = 0 ; i < temp.size() ; i++) { 
         Rule r = temp.elementAt(i) ; 
         quoty = quotedConsequent(r) ; 
         if (quoty != null) break ; 
      } // end for
      if (quoty == null) return null ; // no more
      // otherwise find other quoties
      for (int i = 0 ; i < temp.size() ; i++) { 
         Rule r = temp.elementAt(i) ; 
         Functor f = quotedConsequent(r) ; 
         if (f != null && f.name.equals(quoty.name) && f.arity == quoty.arity) // functor equals
            quoties.add(r) ;
      }   
      // extract quoties from temp
      for (int i = 0 ; i < quoties.size() ; i++) { 
         temp.remove(quoties.elementAt(i)) ;   
      }   
      return quoties ;   // will be null if no more quoties
   }

   // returns null or first quoted predicate
   static Functor quotedConsequent(Rule r) { 
         if (r.consequent.size() == 1 && 
                r.consequent.elementAt(0).size() == 1 &&                     
                   ((Functor)r.consequent.elementAt(0).elementAt(0)).name.startsWith("'") ) 
            return ((Functor)r.consequent.elementAt(0).elementAt(0)) ;
         else return null ; 
  }

/////////////////////   aft folding /////////////////////
   /**
     * Fold the theory using rewriting:
     *  'quoty'(X) => C    (C a conjunction)
     * is used to subtitute C for 'quoty'(X) 
     * in consequents (aft) in all the rules.
      */
   protected static void aftFold(Theory temp) {
      Theory quoties = findQuotedAntecedents(temp) ;  // quoty => a,b,c3
      // use quoties to rewrite all consequences
      for (int i = 0 ; i < temp.size() ; i++)  // each rule
         for (int j = 0 ; j < quoties.size() ; j ++) 
            // rewrite ith rule using each jth quotie
            temp.setElementAt(rewrite(temp.elementAt(i),quoties.elementAt(j)),i) ;
   }

/* e.g.
 true => dom(X), '{((p(X) & q(X)) | ~r(X))}'(X).
 '{((p(X) & q(X)) | ~r(X))}'(X) => '{(p(X) & q(X))}'(X) | '{~r(X)}'(X).
 '{(p(X) & q(X))}'(X) => p(X), q(X).
 '{~r(X)}'(X) => not_r(X).
*/

   // aft rewrite of rule
   static Rule rewrite(Rule r, Rule quoty) { 
      Disjunction d = new Disjunction() ; 
      for (int i = 0 ; i < r.consequent.size() ; i++) 
         d.add(rewrite(r.consequent.elementAt(i),quoty)) ;
      return new Rule(r.antecedent,d) ; 
   } 

   // aft rewrite of rule
   static Conjunction rewrite(Conjunction c, Rule quoty) { 
      Conjunction cc = new Conjunction() ;  
      for (int i = 0 ; i < c.size() ; i++) { 
         Functor f = c.elementAt(i) ; 
         if (f.index == quoty.antecedent.elementAt(0).index)  // N.B. uses .index
            cc.addAll(quoty.consequent.elementAt(0)) ; 
         else cc.add(f) ; 
      }
      return cc ; 
   }

   // quoty => Conjunction
   static Theory findQuotedAntecedents(Theory temp) {
      Theory quoties = new Theory() ;
      for (int i = 0 ; i < temp.size() ; i++) {
         Rule r = temp.elementAt(i) ; 
         if (r.antecedent.size() == 1 &&
               r.antecedent.elementAt(0).name.startsWith("'") &&
                  r.consequent.size() == 1 ) { // conjunctive
            quoties.add(r) ; 
         }
      } 
      Iterator it = quoties.iterator() ; 
      while (it.hasNext()) temp.remove(it.next()) ; // needs indexing for equals
      return quoties ; 
   }

//////////////////////////////////////////////////////////////////////////////////////////

   /** 
     *     0      ...   (kth rule,antecedent.size())
     *   true => ... => --
     */
   protected static Point starts(Theory thy) { 
      for(int k = 0 ; k < thy.size() ; k++) 
         if (thy.elementAt(k).antecedent.size() > 1 ||  
              thy.elementAt(k).consequent.size() > 1  ||
               thy.elementAt(k).consequent.elementAt(0).size() > 1)
            return new Point(k,thy.elementAt(k).antecedent.size())  ; 
      return new Point(thy.size()-1,thy.elementAt(thy.size()-1).antecedent.size()) ; 
   }


   /**
     *  Throw ForeException when translating formal implications.
      */
   protected class ForeException extends RuntimeException {
      WFF wff ; 
      protected ForeException(WFF w) { this.wff = w ; }
   }

///////////////////////////////////////////////////////////////////////////////

   /**
     * List of Variables.
     */
   protected class VarList extends Vector <Variable>  {}  

///////////////////////////////////////////////////////////////////////////////
 

}


