package sam.infer ;      

import java.util.* ; 
import sam.lang.* ; 
import sam.util.* ; 

/**
  *  An ActiveRule is the agent for applying 
  *  a sam.lang.Rule to a Tree using Tree.table.
  *  @version 25 June 2014
  *  @author &copy; John Fisher
   */
public class ActiveRule  { 

   protected final Rule cologRule ;  // the interpreted rule
   protected final int index ;                   // 1-based index
   protected final Fact[][] table ;              // variable bindings
   protected final int[] choice ;                // choice record
   protected int factor ;                        // antecedent factor
   protected final Tree tree ;                   // the search tree
   public Theory theory ; 
   protected final ActiveTheory activeTheory ;   // the containing active theory
   protected int mark ;                          // mark branch point of last application
   protected final boolean splits ;              // is this a disjunctive rule?
   public static int SkolemNumber = 0 ;  // static indexing of Skolem eigenvariables
  
   protected  int A,    // width of antecedent
                  C,    // A + widest consequent
                  U,    // number of universal variables in the antecedent
                  V ;   // total number of variables

   private int reach ; // length of longest conjunction in consequent
   private int width ; // number of splits in consequent of this rule
   protected Functor[] FACTOR ;  // the Functor factors of the antecedent 0..(A-1)
   protected boolean start ; 
   protected Index[] forward ;  // consequent checking
   protected Choice[] bucket ;

   /**
     *  Construct an ActiveRule which knows its
     *  colog rule and the Tree which it builds.
     */
   public ActiveRule(Rule r, Tree dftree, ActiveTheory active, int indx) { 
      this.index = indx ; 
      this.cologRule = r ; 
      this.tree = dftree ;         // dftree.activeTheory NOT YET constructed
      this.theory = tree.theory ;
      this.activeTheory = active ; // passed from its constructor
      this.A = r.A ; 
      this.C = r.C ;    
      this.V = r.V ; 
      this.U = r.U ; // U == 0 means factual 
      this.table =  new Fact[V][C] ; //  [0,V) x [0,C)
      this.choice = new int[C] ; 
      this.mark = -1 ; // not used on current branch
      this.splits = this.cologRule.consequent.size() > 1 ;
      // reach of this rule
      int rch = this.cologRule.consequent.elementAt(0).size() ; 
      for (int k = 1 ; k < this.cologRule.consequent.size() ; k++) {
         int l = this.cologRule.consequent.elementAt(k).size() ; 
         if (l > rch) rch = l ; 
      }
      this.reach = rch ;  
      this.width = this.cologRule.consequent.size() ; 
      FACTOR = new Functor[A] ;
      for (int f=0 ; f<A ; f++)  
         FACTOR[f] = this.cologRule.antecedent.elementAt(f) ;
      this.forward = this.cologRule.forward ; // bring it here 
      this.bucket = new Choice[C] ; 
   }  

////////////////////////////////////////////////////////////////
/*  bindings table, variables down left, rule factors across top

     antecedent [0,A)   consequent [A,C)
     0  1  2  3 ... A-1 A  ... C-1 C     
  0                    |           |
  1                    |           |
  2         []         |    *      |   Universal variable 
  .                    |           |    bindings [0,U) x [0,A)
  .                    |           |
  U-1                  |           |
  ---------------------|-----------|
  U                    |           |   
  .       *            |    []     |   Existential variable
  .                    |           |     bindings [U,V) x [A,C)
  V-1------------------|-----------| 
*///////////////////////////////////////////////////////////////

   // clear entire table, not called, inlined where needed
   public void clear() { 
      for (int a = 0 ; a < V ; a++)     // [0,V)
         for (int b = 0 ; b < C ; b++)  // [0,C)   
            table[a][b] = null ; // unbound
   }

   /**
     * initialize choice 
     * @return false if some factor has no first, else true ; 
      */
   protected boolean initChoice() {
      for (int i = 0 ; i < A ; i++)    {
         int c = activeTheory.first[FACTOR[i].index] ; 
         if (c == tree.depth) return false ;                          
         choice[i] = c ;      
      }  
      return true ; // initialized        
   }                                                              

   //   public int apply() ; // defined in ActiveDCRule

   /**
     *  @return hash code for antecedent factor using key pattern.  
      */
   protected int keyHashCode(Functor f) {
      int h = f.index ;
      for (int i = 0 ; i < f.arity ; i++) 
         h = (h<<5) + (f.key[i] ? hashCode(f.args[i]) : 0) ; 
      return h ; 
   } 

   /**
     *  @return factual hash value of terms using the current table bindings.  
     *  This call is only made for BOUND Terms.
      */
   public int hashCode(Term t) {   
      int hc = 0 ; 
      if (t.isVariable()) { 
         int v = ((Variable)t).index ; // which variable?
         for(int c = 0 ; c < C ; c++) 
            if (table[v][c] != null) 
               hc= Fact.hashcode(table[v][c]) ; 
            // N.B. variable must be bound 
      }
      else {  //t is a Functor 
         hc = ((Functor)t).index ; 
         for (int i = 0 ; i < ((Functor)t).arity ; i++) {
            int h = hashCode(((Functor)t).args[i]) ; 
            if (h == 0) return 0 ; // unbound arg   ???
            hc = (hc << 5)  + h ;     
         }
      } 
      return hc ; 
   }

//-----------------------  match_order_args ----------------------------
   /**
     *  Functor free and Fact fixed already have same index,
     *  so predicates match.  
     *  (22 December 2011 tests: about same speed as before.)
     *  Match arguments using factor orders.
     *  @return whether match can be made.
     */
   public boolean match_order_args(int n, Functor free, Fact fixed) { 
      for(int i=0; i < free.arity ; i++) { 
         int o = free.order[i] ; // use free's order 
         if (!matchTerm(n, free.args[o], fixed.args[o])) return false ; 
      }
      // terms all matched
      return true ; 
   }
//----------------------------------------------------------------------

   /**
     *  Functor free and Fact fixed already have same index,
     *  so predicates match.
     *  Attempt to match arguments using bindings. 
     *  @return whether match can be made.
     */
   public boolean match_args(int n, Functor free, Fact fixed) { 
      for(int i=0; i < free.arity ; i++) // use left-to-right order
         if (!matchTerm(n, free.args[i], fixed.args[i])) return false ; 
      // terms all matched
      return true ; 
   }

   /**
     *  Match term.
     *  @arg n is the factor (0..A-1 or A..C) at which a binding could be made. 
     *  @return whether match can be made. 
     */
   public boolean matchTerm(int n, Term free, Fact fixed) {
      if (free instanceof Variable) {  
         // Compute the value of the variable -- "lookup"
         int d = ((Variable)free).index ; // which variable
         Fact f = null ;
         for(int c=0 ; c <= n ; c++) 
            if (table[d][c] != null) { 
               f = table[d][c] ; 
               break ; 
         }
         // f is the lookup value (or null)
         if (f != null) {    // Fact entry                       
            if (f.index == fixed.index) {
               if (f.args != null) {
                  for(int i=0; i < f.args.length ; i++) 
                     if (!matchTerm(n, f.args[i], fixed.args[i])) return false ; 
               }
               // otherwise no terms or all matched
               return true ; 
            }
            else return false ; 
         }
         else { 
            table[d][n] = fixed ;  // bind dth variable to fixed Fact 
            return true ; 
         }
      }  // otherwise free is Functor or Fact
      else if (free.index != fixed.index) return false ; 
      else if (free instanceof Functor) {
         for(int i=0; i < ((Functor)free).args.length ; i++) 
            if (!matchTerm(n, ((Functor)free).args[i], fixed.args[i])) return false ; 
         // terms all matched
         return true ; 
      }
      else {  // free is a Fact
         if (((Fact)free).args != null) {
            for(int i=0; i < ((Fact)free).args.length ; i++) 
               if (!matchTerm(n, ((Fact)free).args[i], fixed.args[i])) return false ; 
         }
         // terms all matched
         return true ; 
      }
   }

   /**
     *  Is the consequent already subsumed on this branch
     *  or is the consequent too complex?
     *  Bindings were determined to focus. 
     */
   protected boolean subsumed() { 
      if (tooComplex(tree.complexity_limit)) return true ;  // subsumed by complexity!
      Disjunction consequent = cologRule.consequent ; 
      boolean alreadySATISFIED = false ; 
      for(int i = 0 ; i < consequent.size() ; i++) { 
         // try to satisfy some conjunct  
         Conjunction con = consequent.elementAt(i) ;   
         ///// reset table for existential binding
         for (int v = U ; v < V ; v++)     // [U,V)
            for (int c = A ; c < C ; c++)  // [A,C)
               table[v][c] = null ;  
         ///// reset consequent portion of choice, bucket
         for (int j = 0 ; j < con.size() ; j++) {              
            Functor f = (Functor)con.elementAt(j) ;           
            choice[A+j] = activeTheory.first[f.index] ;  
            //if (f.subfactual) //  BUG,  too soon
            //   bucket[A+j] = tree.table.get(keyHashCode(f)) ;     
            bucket[A+j] = null ; // -> signals to initialize after previous factor bound 
         }                                                    
         /////   
         if (satisfied((Conjunction)consequent.elementAt(i)) ) {
            alreadySATISFIED = true ;  
            break ;
         } 
      }
      if (alreadySATISFIED) tree.subsumed++ ; 
      return alreadySATISFIED ;
   }

    /**
     *  Is a conjunction in the consequent already satisfied?
     *  Existential variables may be temporarily bound. 
     *  Factual factors are checked via tree.table.
      */
   protected boolean satisfied(Conjunction c) { 
      int k = 0 ;  // 0th conjunct
      while (k != -1) {  // k loop, until true or cannot be satisfied  
         if ( k == c.size() ) {  
            return true ; // all conjuncts already satisfied  
         } 
         else {  // probe
            boolean matched = false ;
            Functor f = (Functor)c.elementAt(k) ;    
            // FACTUAL factor
            if (f.factual)  matched = (factual(A+k,f)) ; // sets choice[A+k] ;-)
            // SUBFACTUAL factor
            else if (f.subfactual) { // attempt to find a bucket match 
               // -> fill bucket using any previous bindings
               if (bucket[A+k] == null) bucket[A+k] = tree.table.get(keyHashCode(f)) ; 
               // otherwise backtracking 
               while (bucket[A+k] != null) {  // "bucket loop"
                  if (bucket[A+k].factual) // move past bucket factual
                     bucket[A+k] = bucket[A+k].next ; // stops bucket loop if null
                  else {   // attemp to match subfactual entry
                     if (match_args(A+k,f,tree.branch[bucket[A+k].index])) {
                        matched = true ;
                        choice[A+k] = bucket[A+k].index ; // where matched in the tree
                        break ;  // break the bucket loop
                     }  
                     else {  // undo any partial bindings from match attempt
                        ///// nullify (A+k)th column /////
                        for(int v = U ; v < V ; v++) 
                           table[v][A+k] = null ;     
                        // move to next in bucket
                        bucket[A+k] = bucket[A+k].next ; 
                     }
                  }  
               }         
            } // end bucket probe                                
            else { // LINEAR factor
               int p = choice[A+k] ;  // start at backtrack point in the tree 
               // travel down tree to attempt match
               //boolean matched = false ; 
               while (p <= tree.focus) {    
                     if (match_args(A+k,f,tree.branch[p])) {
                        matched = true ;
                        choice[A+k] = p ; // where matched in the tree
                        break ; // breach the linear loop
                     }
                     else {  // undo any partial bindings
                        ///// nullify (A+k)th column /////
                        for(int v = U ; v < V ; v++) 
                           table[v][A+k] = null ;     
                        /////
                        p = activeTheory.jumpto[p] ;   
                     }
               }  // end while
            } // end linear probe
            //////////////// matched or BACKTRAK//////////////////////////////////
            if (matched)  k++ ; // next factor
            else { // BACKTRACK 
               ///////////////////////////////////////////////////////////////////
               // Backtracking moves "forward" -> for subfactual or linear choices
               // The "back" part is going "left" <- on the factors.  
               ///////////////////////////////////////////////////////////////////
               while (true) {  // <- on factors
                  ///// <- past factuals
                  if (((Functor)c.elementAt(k)).factual) {
                     while (k != -1 && ((Functor)c.elementAt(k)).factual)    
                        k-- ;  // <-
                     if (k == -1) return false ;  
                  }
                  for(int u = U ; u < V ; u++)  // clear column A+k of table
                     table[u][A+k] = null ;  
                  /////  
                  if (((Functor)c.elementAt(k)).subfactual) { 
                     if (bucket[A+k] != null)  bucket[A+k] = bucket[A+k].next ;   
                     if (bucket[A+k] == null) {  // move <- on factor
                        k-- ;  
                        if (k == -1) return false ;  
                     }  
                     else { // move -> in bucket & bucket nonempty
                        break  ;   
                     }
                  }  // subfactual
                  else { // linear factor    
                     int nxt = tree.depth ; 
                     if (choice[A+k] <= tree.focus) // last attempt within range
                        nxt = activeTheory.jumpto[choice[A+k]] ; 
                     if (nxt <= tree.focus) {  
                        choice[A+k] = nxt ; 
                        break ; // backtrack loop
                     }
                     else { // more backtracking 
                        choice[A+k] = 
                           activeTheory.first[((Functor)c.elementAt(k)).index] ; 
                        k-- ; // <-
                        if (k == -1) return false ;  
                     }
                  }  // linear
               } // end while
            }// end BACKTRACK  
         } // end probe
      } // end  <- k loop
      return false ; // not already satisfied
   }

///////////////////////////////////////////////////////////////////////
///////  Use tree.table to compute subsumption-by-fact-in-tree
//////////////////////////////////////////////////////////////////////
   /**
     *   Check tree.Table for previous fact on branch.
     *   If match and n < A set choice[n] branch match index. 
      */
   protected boolean factual(int n, Functor f) {
      int h = this.hashCode(f) ; 
      Choice c = tree.table.get(h) ;   
      if (c == null) return false ;
      //////////Iterator it = index.iterator() ;
      while (c != null) { 
         if (!c.factual) c = c.next ;
         else {  // factual entry but may not match
            int b = c.index ; // branch position
            Fact fxd = tree.branch[b] ;
            if (  // could check hash value ???
                  f.index == fxd.index  &&       // same predicate
                    match_order_args(n,f,fxd) ) {  // arguments match
               if (n < this.A) this.choice[n] = b ;   // antecedent only
               return true ; 
            } 
            c = c.next ;
         }
      } 
      return false ;
   }

//////////////////////////////////////////////////////////////////////

   /**
     *  Expand downward at focus using
     *  the disjunctive consequences of this rule 
     *  and bindings from table.  
     *  First disjunct determines continuation on tree.branch, others saved.
     */
   public int expand() { 
      // check bounds ... These block branch
      //if (this.width - 1 + tree.branchCount > tree.width) {
      //if (this.width - 1 + tree.numberOfBranches() > tree.width) {
      if (this.width  + tree.numberOfBranches() > tree.width) {
         if (tree.log) {  // record which inference would have been made
            tree.rule[tree.count] = this.index ; 
            tree.inference[tree.count] = 
               new Inference(substitute(cologRule.antecedent),substitute(cologRule.consequent)) ;
         }
         return -2 ;  // too wide
      }
      if (this.reach + tree.focus + 1 > tree.depth) {
         if (tree.log) {  // record which inference would have been made
            tree.rule[tree.count] = this.index ; 
            tree.inference[tree.count] = 
               new Inference(substitute(cologRule.antecedent),substitute(cologRule.consequent)) ;
         }
         return -1 ;  // too deep
      }
      ///////////// 
      // otherwise, just right ...
      // Construct antecedent from backtrack record
      And ant = new And(this.A) ;
      for (int i = 0 ; i < this.A ; i++)  {
         ant.add(tree.branch[choice[i]]) ; 
         ant.choice[i] = choice[i] ;
      }
      tree.ant[tree.focus] = ant ;   ///  RECORD FOR RELEVANCY COMPUTATION 
      And[] conseq = substitute(cologRule.consequent) ; 
      if (conseq.length > 1) {  
         tree.choices.push(conseq) ; 
         tree.fstack.push(tree.focus) ; 
         ///////////////////////////////////////////////////////////////////
         tree.qstack.push(activeTheory.rrQ.next) ;     //  save rrQ.next for branches
         //activeTheory.eQ = activeTheory.eQ.clone() ; // use clone for branch
         //////////////////////////////////////////////////////////////////
         tree.next.push(1) ; // the next choice 1-based
         if (tree.log) tree.cstack.push(tree.count) ;  // save attach point
         //tree.branchCount += this.width - 1 ;
      }
      if (tree.log) {  
         //And ant = new And(this.A) ;
         //for (int i = 0 ; i < this.A ; i++) ant.add(tree.branch[choice[i]]) ; 
         tree.rule[tree.count] = this.index ; 
         tree.inference[tree.count] = 
            new Inference(ant,conseq) ;
      }
      And c = conseq[0] ; 
      int save_focus = tree.focus ; // count at start of add
      for (int i = 0 ; i < c.size() ; i++) { 
         Fact f = c.elementAt(i) ; 
         //f.focus = save_focus ; // f was inferred at save_focus
         // if (!tree.occurs(f)) {  
            tree.focus++ ;  
            tree.count++ ; 
            tree.branch[tree.focus] = f ; // depth overflow guarded  
            tree.inferredAt[tree.focus] = save_focus ;
///// >>>
            tree.remember(f) ; // remember
///// <<<
            // adjust indexing
            int frst = activeTheory.first[f.index] ;
            if (frst == tree.depth)   // none yet
               activeTheory.first[f.index] = tree.focus ;
            else  // jump from last to here  
               activeTheory.jumpto[activeTheory.last[f.index]] = tree.focus ;
            activeTheory.last[f.index] = tree.focus ;  // last is here
            activeTheory.jumpto[tree.focus] = tree.depth ;   // jump off branch
            /////
            if (tree.log) {   
               tree.fact[tree.count] = f ;
               tree.parent[tree.count] = tree.count-1 ;   
               // set where used
               // f.focus = currentcount ; 
            }  
          //} // end if occurs check
      }
      return tree.focus ;  
   } // end expand() 


   /*** Substitute into a Disjunction.*/
   public And[] substitute(Disjunction d) { 
      And[] d_new = new And[d.size()] ; 
      for(int i = 0 ; i < d.size() ; i++)
         d_new[i] = substitute( (Conjunction)d.elementAt(i) ) ; 
      return d_new ; 
   }   

   /*** Substitute into a Conjunction.*/
   public And substitute(Conjunction c) { 
      And ant = new And(c.size()) ; 
      ///// reset existential variables
      for (int v = U ; v < V ; v++)     // [U,V)
         for (int f = A ; f < C ; f++)  // [A,C)   
            table[v][f] = null ; 
      /////      
      for(int i = 0 ; i < c.size() ; i++)
         ant.add(substitute((Functor)c.elementAt(i))) ; 
      return ant ; 
   }   

   /**
     *  Substitute into a Term. 
     *  Note that a Skolem constant Fact is 
     *  a special case.  
     *  Also sets hash code for Fact.
     */
   public Fact substitute(Term t) {
      if (t.isVariable()) { 
         // Compute the value of the variable
         int v = ((Variable)t).index ; // which variable?
         Fact f = null ;
         for(int c=0 ; c <= A ; c++) 
            if (table[v][c] != null) { 
               f = table[v][c] ; 
               break ; 
         }
         // f is the lookup value (or null)
         if (f == null) {    // must be new Skolem
            Fact sk =   //         # of normal Facts         +  Skolem#
               new Fact(theory.name.length+SkolemNumber,0) ; // N.B. index
            table[v][A] = sk ;
            SkolemNumber++ ; 
            return sk ;    // sk has hash code set = its index
         }
         else return f ;  // f is already set
      }  // t was  Variable
      else { // t is Functor
         Fact f_new = new Fact(((Functor)t).index,((Functor)t).arity) ; 
         if (f_new.args != null) {
            for(int i=0 ; i < f_new.args.length ; i++)
               f_new.args[i] = substitute(((Functor)t).args[i]) ; 
         }
         ///////// 22 April 2013
//         f_new.setHashT() ;  // arguments have hash code set
         /////////
         return f_new ; 
      }
   }

////////////////////////////////////////////////////////////
///////  Dynamic complexity of expressions -- depth of operators  
//////      applied only for a complex rule
///////////////////////////////////////////////////////////

   // called by subsumption
   // over-ridden in ActiveFactualDCRule
   protected boolean tooComplex(int limit) { 
      if (!this.cologRule.complex)  return false  ;  // experiment 10 March 2013, only slight speedup (?)
      if (limit == -1) return false ;   // no limit, so do not compute
      // otherwise
      for(int i = 0 ; i < this.cologRule.consequent.size() ; i++) 
         if (tooComplex(this.cologRule.consequent.elementAt(i),limit)) return true ;
      return false ; 
   }

   protected boolean tooComplex(Conjunction c,int limit) { 
      for(int i = 0 ; i < c.size() ; i++) 
         if (tooComplex(c.elementAt(i),limit)) return true ;
      return false ; 
   }
  
   /*** Does predicate have too complex an argument? */
   protected boolean tooComplex(Functor f,int limit) { 
      for(int i = 0 ; i < f.args.length ; i++)  
         if (termComplexity(f.args[i]) > limit) return true ; 
      return false ;
   }  

   /** 
     *  Term complexity of expresssion in rule,   
     *  total number of operations, computed using lookup.
     *  Complexity taken to be maximum depth, 
     *  so that algebraic rearrangements have same complexity. 
     */
   protected int termComplexity(Term t) {
      if (t.isVariable())  { // look it up in the VarBind
         Fact f = this.lookup(((Variable)t).index) ;  // lookup Fact
         if (f == null) return 0 ;     // eigenvariable
         else return factComplexity(f) ; 
      }
      // otherwise t is a Functor 
      else if (((Functor)t).arity == 0) return 0 ; // a constant in rule
      else { 
         int c = 0 ; 
         for (int j=0 ; j < ((Functor)t).arity ; j++) 
            c +=  termComplexity(((Functor)t).args[j])  ; 
         return 1+c ; // total number of operators
      }
   }

   /*** Fact complexity is total number of opertors in Fact */   
   protected int factComplexity(Fact t) {  //SEEMS CORRECT
      if (t.args == null) return 0 ; 
      // otherwise f(.?.) has complexity 1 + ?
      int c = 0 ; 
      for (int j=0 ; j < t.args.length ; j++) 
         c+=  factComplexity(t.args[j])  ; 
      return 1+c ; 
   }

   /**
     *  @return table value of variable/index.
      */
   protected Fact lookup(int var) {
      for(int k=0 ; k < this.C ; k++) 
         if (table[var][k] != null)  
            return table[var][k] ; 
      return null ; // would be eigenvariable
   }

/////////////////////////////////////////////////////////////////////////
/* NOT CALLED BY OTHERS.  Reinstate if needed.
   public boolean match(int n, Functor free, Fact fixed) { 
      if (free.index == fixed.index) { 
         if (free.arity == 0) return true ; 
         else {   
            for(int i=0; i < free.arity ; i++) 
               if (!this.matchTerm(n, free.args[i], fixed.args[i])) return false ; 
            // terms all matched
            return true ; 
         }
      }
      else return false ; 
   }
*/

   private static void sopl(Object o) { System.out.println(o) ; }
}

