package sam.lang ;

import java.util.* ;
import java.io.* ; 
import sam.lang.* ;
import sam.util.* ;

/**
  *  Compute static properties of a colog theory:
  *     index rule variables, set U and V 'field'
  *     initialize predicate 'key' and set values 
  *     set Rule 'forward' fields  
  *     set predicate 'factual' field and rule 'factual' field
  *     set predicate match 'order' 
  *     set key patterns for theory 
  *     (Uses static inner classes for generics. ;-)
  *        started March 2014, Mammoth Lakes, CA
   *       last 14 June 2014
    */
public class CologAnalyzer {

   // non-generic static renaming, can use for [] and casting, w/o warnings
   static class VarSet  extends HashSet<Variable> {} 
   static class PatternVector  extends Vector<Pattern> {} 
   static class VarMap  extends HashMap<Variable,Integer> {} 

   /**
     *  Analyze colog theory.
     *  call like this from commline:
     *  java -cp . sam.lang.CologAnalyzer <_full_path_file_name> ;
      */
   public static void analyze(Theory theory) { 
      trimToSize(theory) ;
      indexFunctors(theory) ; 
      Iterator it = theory.iterator() ;
      while (it.hasNext())  {
         Rule r = (Rule)it.next() ;
         Index[] forward = indexRuleVariables(r) ; 
         r.forward = forward ;
         setOrderAndFactual(r) ; 
         r.complex = complex(r) ; 
      }
      theory.keyPatterns = setKeyPatterns(theory) ;  
      // Determine whether any key pattern for theory is subfactual
      theory.subfactual = anySubFactual(theory.keyPatterns) ; 
   }

   // @return whether any predicate pattern is subfactual
   private static boolean anySubFactual(Vector<Pattern>[] patterns) {
      for (int p = 0 ; p < patterns.length ; p++) { 
         Vector<Pattern> v = patterns[p] ; 
         if (v != null) { // for a predicate
            Iterator itr = v.iterator() ; 
            while (itr.hasNext())  
               if (((Pattern)itr.next()).subfactual) return true ; 
         }
      }
      return false ; 
   }

   /**
     *  Compute index for functors (predicates and operators) 
     *  of this theory, and set index field for functors. 
     *  Should call AFTER all rules are added.  Creates 
     *  name[] and type[] for theory.
      */
   protected static void indexFunctors(Theory theory) {
      // functor name/arity --> index 
      HashMap<String,Integer> indexmap = new HashMap<String,Integer>() ; 
      indexmap.put("goal/0",Theory.GOAL_INDEX) ;    // 1
      indexmap.put("false/0",Theory.FALSE_INDEX) ;  // 2
      indexmap.put("true/0",Theory.TRUE_INDEX) ;    // 0
      // functor name/arity --> type 
      HashMap<String,Byte> typemap = new HashMap<String,Byte>() ;
      typemap.put("goal/0",Functor.FCNL) ; 
      typemap.put("false/0",Functor.FCNL) ; 
      typemap.put("true/0",Functor.FCNL) ; 
      Iterator it = theory.iterator() ; 
      theory.count = 3 ; 
      while (it.hasNext()) { 
         Rule r = (Rule)it.next() ; 
         Conjunction a = r.antecedent ; 
         Iterator ait = a.iterator() ; 
         while(ait.hasNext()) { 
            Functor f = (Functor)ait.next() ; 
               indexFunctors(f,theory,indexmap,typemap) ;             
         }
         Disjunction d = r.consequent ; 
         Iterator dit = d.iterator() ; 
         while (dit.hasNext()) { 
            Conjunction c = (Conjunction)dit.next() ;
            Iterator cit = c.iterator() ; 
            while (cit.hasNext()) { 
               Functor f = (Functor)cit.next() ; 
               indexFunctors(f,theory,indexmap,typemap) ; 
            }
         }
      }
      // now all functors are mapped, so create name[], type
      theory.name = new String[indexmap.size()] ; 
      theory.type = new byte[indexmap.size()] ; 
      Iterator fcns = indexmap.keySet().iterator() ; 
      while (fcns.hasNext()) { 
         String s = (String)fcns.next() ; 
         int i = indexmap.get(s) ; 
         byte t = typemap.get(s) ;
         int k = s.indexOf("/") ; 
         s = s.substring(0,k) ; 
         theory.name[i] = s ; 
         theory.type[i] = t ;
      }
   }


   // map indexes for functor and arguments
   protected static void indexFunctors(Functor f,Theory thy,HashMap<String,Integer> imap,HashMap<String,Byte> tmap) { 
      Object probe = imap.get(f.name+"/"+f.arity) ; 
      if (probe != null) 
         f.index = (Integer)probe ; 
      else {
         f.index = thy.count ; 
         imap.put(f.name+"/"+f.arity,thy.count) ; 
         thy.count++ ;
      }      
      tmap.put(f.name+"/"+f.arity,f.type) ;
      // also do for functor args
      for (int i = 0 ; i < f.arity ; i++) {
         Term t = f.args[i] ; 
         if (t instanceof sam.lang.Functor) 
            indexFunctors((Functor)t,thy,imap,tmap) ; 
      }
   }

   /**
     *  Index RULE variables.
     *  Automatically sets predicate 'key' patterns
     *  for uniform tableing, predicate match 'order'
     *  and predicate 'factual' field.
     *  @return the forward lists for r
      */
   protected static Index[] indexRuleVariables(Rule r) { 
      Index[] forward = new Index[r.antecedent.size()] ; 
      for (int j = 0 ; j < r.antecedent.size() ; j ++) 
         forward[j] = new Index() ; 
      VarMap Umap = new VarMap() ;
      VarSet[] Uvars = indexAntecedentVariables(r.antecedent,Umap) ; 
      r.U = Umap.size() ; 
      Disjunction d = r.consequent ; 
      int s = 0 ;
      VarMap tmp = new VarMap() ;
      for(int i = 0 ; i < d.size() ; i++) { 
         tmp.clear() ; 
         tmp.putAll(Umap) ;  // clone warns
         VarSet vs = indexConsequentVariables(d.elementAt(i),tmp,Umap.keySet().size()) ;  // BREAK IN HERE
         if (tmp.size() > s) s = tmp.size() ; 
         forward[forward(Uvars,vs)].add(i) ;
      }   
      r.V = s ; 
      r.existential = r.V - r.U ; 
      return forward ;
   }

   /**
     *  Index the variables inside ANTECEDENT terms.
     *  Automatically initializes and sets antecedent predicate key.
      */
   protected static void  indexAntecedentVariables(Term t, VarMap map) { 
      if (t instanceof Variable) { 
         Object ob = map.get((Variable)t) ;
         if (ob != null) ((Variable)t).index = (Integer)ob ;  
         else { 
            ((Variable)t).index = map.size() ; 
            map.put((Variable)t,map.size()) ;
         }
      }
      else if (t instanceof Functor) {
         Functor f = (Functor)t ; 
         VarMap save_map = (VarMap)map.clone() ;     //__NB_// 
         for(int i = 0 ; i < f.arity ; i++) {  
            indexAntecedentVariables(f.args[i], map) ;   
            if (f.key != null) { // f is a predicate, so set f.key[i]
               VarSet vs = new VarSet() ;                                 //__NB_//
               findVars(f.args[i],vs) ;                                   //__NB_//
               f.key[i] = (((VarMap)save_map).keySet()).containsAll(vs) ; //__NB_//
            }
         }
      }
      else {}
   }   

   // Compute the VarSet for a predicate factor.                          //__NB_//
   private static void findVars(Term t, VarSet set) { 
      if (t instanceof Variable)  set.add((Variable)t) ; 
      else { 
         Functor f = (Functor)t ;
         for (int i = 0 ; i < f.arity ; i++) 
            findVars(f.args[i],set) ; 
      }
   }

   /**
     * Public findVars for sam.lang.fol.WFF to call 
      */
   public static Set findVars(Term t) { 
      VarSet vars = new VarSet() ;
      findVars(t,vars) ;  
      return vars ;
   }

   /**
     *  Index variables in ANTECEDENT conjunction 
     *  @return cumulative variable sets for factors, left-to-right
      */
   protected static VarSet[] indexAntecedentVariables(Conjunction c, VarMap map) { 
      VarSet[] set = new VarSet[c.size()] ;  
      for(int i = 0 ; i < c.size() ; i++) { 
         Functor f = (Functor)c.elementAt(i) ; 
         f.key = new boolean[f.arity] ;  // !=null signals f is predicate
         indexAntecedentVariables(f, map) ; 
         set[i] = new VarSet() ; 
         set[i].addAll(map.keySet()) ; 
      }
      return set ; 
   }

   /**
     *  Index variables in a CONSEQUENT term.
     *  map is index map of variables, to which existential variables are also put.
     *  U is the number of universal variables in the antecedent .
     *  @return set of universal variables in this consequent conjunction
      */
   protected static VarSet indexConsequentVariables(Term t,VarMap map, int U) { 
      VarSet set = new VarSet() ;
      if (t instanceof Variable) { 
         Object ob = map.get((Variable)t) ;
         if (ob != null) { 
            ((Variable)t).index = (Integer)ob ;  
            if ((Integer)ob < U) set.add((Variable)t) ; // a universal variable
         }
         else { 
            ((Variable)t).index = map.size() ; 
            map.put((Variable)t,map.size()) ;
         }
      }
      else if (t instanceof Functor) { 
         Functor f = (Functor)t ; 
         VarMap save_map = (VarMap)map.clone() ;                       //__NB_// 
         for(int i = 0 ; i < f.arity ; i++) {      
            VarSet uvars = indexConsequentVariables(f.args[i],map,U) ;
            set.addAll(uvars) ;          
            if (f.key != null) { // f is a predicate, so set f.key[i]
               VarSet vs = new VarSet() ;                                 //__NB_//
               findVars(f.args[i],vs) ;                                   //__NB_//
               f.key[i] = (((VarMap)save_map).keySet()).containsAll(vs) ; //__NB_//
           }
         }
      }
      else {}
      return set ;
   }   

   /**
     *  Index variables in CONSEQUENT conjunction 
     *  @return set of universal variables in the conjunction
      */
   protected static VarSet indexConsequentVariables(Conjunction c,VarMap map,int U) { 
      VarSet set = new VarSet() ;  
      for(int i = 0 ; i < c.size() ; i++) { 
         Functor f = (Functor)c.elementAt(i) ; 
         f.key = new boolean[f.arity] ;  // !=null signals f is predicate, needs key
         set.addAll(indexConsequentVariables(f,map,U)) ; 
      }
      return set ; 
   }  

   /**
     *  @return smallest i such that antsets[i] contains all 
     *  universal variables from dset.
      */
   protected static int forward(VarSet[] antsets,VarSet dset) { 
      int fwrd = 0 , max = antsets[antsets.length-1].size() ;  
      Iterator it = dset.iterator() ; 
      while(it.hasNext()) {
         Variable v = (Variable)it.next() ; 
         if (v.index < max) {
            for(int i = 0 ; i < antsets.length ; i++) { 
               if (antsets[i].contains(v)) {
                  if (i > fwrd) fwrd = i ; 
                  break ; // smallest i for this v
               }
            }
         }
      }
      return fwrd ; 
   }  

   /**
     *  Use predicate key to set order and factual
      */
   protected static void setOrderAndFactual(Rule r) {
      Iterator ant = r.antecedent.iterator() ;
      while (ant.hasNext()) {
         Functor f = (Functor)ant.next() ;
         f.order = order(f.key) ;
         f.factual = factual(f.key) ;
         if (f.factual) r.factual = true ;  // antecedent factual
         f.subfactual = subfactual(f.key) ;
         if (f.subfactual) r.subfactual = true ;   
      }
      Iterator d = r.consequent.iterator() ; 
      while (d.hasNext()) {
         Conjunction conj = (Conjunction)d.next() ;
         Iterator c = conj.iterator() ; 
         boolean factual_flag = true ; 
         while (c.hasNext()) { 
            Functor f = (Functor)c.next() ;
            f.order= order(f.key) ;
            f.factual = factual(f.key) ;  // ??? c.factual 
            if (!f.factual) factual_flag = false ;
            f.subfactual = subfactual(f.key) ;
         }
         conj.factual = factual_flag ; 
      }
   }

   /**
     *  @return order int[] based on boolean[] key 
      */
   protected static int[] order(boolean[] key) { 
      int[] ord = new int[key.length] ; 
      int k = 0 ; 
      for (int i = 0 ; i < key.length ; i++) {
         if (key[i]) { ord[k] = i ; k++ ; } 
      }
      for (int i = 0 ; i < key.length ; i++) {
         if (!key[i]) {ord[k] = i ; k++ ; } 
      }
      return ord ;
   }

   /**
     *  @return whether boolean[] key is factual (all true)
      */
   protected static boolean factual(boolean[] key) {
      for (int i = 0 ; i < key.length ; i++) 
         if (!key[i]) return false ;
      return true ; 
  }

   /**
     *  @return whether boolean[] key is subfactual 
      */
   protected static boolean subfactual(boolean[] key) {
      int fc = 0 ;
      for (int i = 0 ; i < key.length ; i++) 
         if (!key[i]) fc++ ;
      if (fc == 0) return false ;  // no unbound argument
      else if (fc <  key.length)  return true ; // some unbound arg
      else return false ;
  }

   /**
     *  for all rules,
     *  for all functors in rule,
     *  record key patterns.
     *  Need to call functor indexing first.  NULL exceptions
      */
   protected static PatternVector[] setKeyPatterns(Theory theory) {
      PatternVector[] keyPatterns = new PatternVector[theory.name.length] ; 
      Iterator rules = theory.iterator() ; 
      while (rules.hasNext()) { 
         Rule r = (Rule)rules.next() ;
         Iterator ants = r.antecedent.iterator() ;
         while (ants.hasNext()) { 
            Functor f = (Functor)ants.next() ; 
            Pattern pat = new Pattern(f.key) ;  
            if (keyPatterns[f.index] ==null) keyPatterns[f.index] = new PatternVector() ;
            if (!keyPatterns[f.index].contains(pat)) keyPatterns[f.index].add(pat) ; 
         }
         Iterator cons = r.consequent.iterator() ; 
         while (cons.hasNext()) {
           Conjunction c = (Conjunction)cons.next() ; 
           Iterator cit = c.iterator() ;
           while (cit.hasNext()) { 
              Functor f = (Functor)cit.next() ; 
              Pattern pat = new Pattern(f.key) ; 
              if (keyPatterns[f.index] ==null) keyPatterns[f.index] = new PatternVector() ;
              if (!keyPatterns[f.index].contains(pat)) keyPatterns[f.index].add(pat) ; 
           }
         }
      }
      // trim pattern vectors to size here
      for(int i = 0 ; i < keyPatterns.length ; i++)
         if(keyPatterns[i] != null)  keyPatterns[i].trimToSize() ;
      return keyPatterns ; 
   }

   // rules have fixed data so trim to size
   private static void trimToSize(Theory theory) { 
      theory.trimToSize() ; 
      Iterator rit = theory.iterator() ;
      while (rit.hasNext()) {
         Rule r = (Rule)rit.next() ; 
         r.antecedent.trimToSize() ; 
         r.consequent.trimToSize() ; 
         Iterator cit = r.consequent.iterator() ;  
         while (cit.hasNext()) ((Conjunction)cit.next()).trimToSize() ;
      }
   }

////////////////////////////////////////////
///   COMPLEXITY calculation methods
////////////////////////////////////////////

   // Complexity of disjunction is max complexity of disjuncts 
   protected static int complexity(Disjunction disj) {
      int c = 0 ;
      for(int i = 0 ; i < disj.size() ; i++) {
         int k =  complexity(disj.elementAt(i) ) ; 
         if (k > c) c = k ; 
      }
      return c ; 
   }

   // Complexity of conjunction is max complexity of conjuncts
   protected static int complexity(Conjunction conj) {
      int c = 0 ;
      for(int i = 0 ; i < conj.size() ; i++) {
         int k =  complexity(conj.elementAt(i) ) ; 
         if (k > c) c = k ; 
      }
      return c ; 
   }

   // Complexity of functor is max complexity of its terms 
   protected static int complexity(Functor f) {
      int c = 0 ; 
      for (int j=0 ; j < f.args.length ; j++) {
         int k =  termComplexity(f.args[j]) ; 
         if (k > c) c = k ;
      }
      return c;
   }

   // static term complexity 
   protected static int termComplexity(Term t) {
      if (t.isVariable())  return 0 ; 
      else if (((Functor)t).arity == 0) return 0 ; 
      else { 
         int c = 0 ; 
         for (int j=0 ; j < ((Functor)t).args.length ; j++) {
            // total number of operations 
            // p(g(f(a,b)))  =>  p(f(g(a),g(b))) is complex
            //c += termComplexity(((Functor)t).args[j]) ; 
            // alternate is DEPTH of operations ..
            // p(g(f(a,b)))  =>  p(f(g(a),g(b))) is NOT complex
            int k =  termComplexity(((Functor)t).args[j])  ; 
            if (k > c) c = k ;
         }
         return 1+c ; 
      }
   }

   /*** Is rule complex? */
   public static boolean complex(Rule r) { 
      return complexity(r.antecedent) < complexity(r.consequent) ;
   }
/////
// end complexity calculations
/////

////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
   /**
     *  Tests, e.g., 
     *     java -cp . sam.lang.CologAnalyzer /colog14A/co/dpe.gl
      */
   public static void main(String[] args) { 
      //String thy = "/Users/jrfisher/Desktop/tst.cl" ; 
      //String thy = "/Colog14A/co/dpe.gl" ; 
      String thy = args[0].trim() ; 
      Theory theory = null ;
      try { 
         sam.lang.antlr.cologLoader loader = 
            new sam.lang.antlr.cologLoader(new FileInputStream(new File(thy))) ; 
         theory = loader.theory ;
      } catch(Exception xx) { System.out.println(xx) ; }
      
      analyze(theory) ;
      System.out.println(args[0] + "\n\n") ; 
      System.out.println(report(theory)) ;  
      /////  complexity 
        //  Functor f = theory.elementAt(0).antecedent.elementAt(0) ; 
        //  System.out.print(f+" complexity="+complexity(f)) ;
   }
////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////// 
////////////////////////////////////////////////////////////////////////////////////
 
   // string shows functor names fields
   private static String inspectNames(Theory theory) { 
      StringBuffer sb = new StringBuffer() ; 
      for (int i = 0 ; i < theory.name.length ; i++) { 
         sb.append(i+" : " + theory.name[i] + "   ") ; 
         sb.append(string(theory.keyPatterns[i])) ;
         sb.append("\n") ; 
      }
      return sb.toString() ;
   }

   /*** @return Theory analysis report */
   public static String report(Theory theory) {
      StringBuffer bf = new StringBuffer() ;  
      bf.append("Functors & predicates by index\n") ; 
      bf.append(inspectNames(theory)) ; 
      bf.append("theory.subfactual="+theory.subfactual) ; 
      bf.append("\nRule attributes\n") ; 
      Iterator thy = theory.iterator() ;
      int n = 1 ; // number the rules, 1-based
      while(thy.hasNext()) {
         Rule r = (Rule)thy.next() ;
         bf.append(n+". "+inspect(r)) ; 
         n++ ;
      } 
      return bf.toString() ; 
    }

   // string shows rule fields
   private static StringBuffer inspect(Rule r) { 
      String spc = " ", nl = "\n" ;
      StringBuffer sb = new StringBuffer() ; 
      sb.append(r.toString()) ; sb.append(nl) ;
      sb.append(">>> A="+r.A+" C="+r.C+" U="+r.U+" V="+r.V + " existential="+r.existential+ " factual="+r.factual+ " subfactual="+r.subfactual+ " complex="+r.complex+"\n") ; 
      Conjunction c = r.antecedent ;
      for (int i = 0 ; i < c.size() ; i++) {
         sb.append(spc) ; sb.append(spc) ; sb.append(spc) ;
         Functor f = (Functor)c.elementAt(i) ; 
         sb.append(i+spc+"key="+string(f.key)) ;
         sb.append(spc) ;
         sb.append(spc+"forward="+r.forward[i].toString()) ;
         sb.append(spc) ;
         sb.append(spc+"order="+string(f.order)) ;
         sb.append(spc) ;
         sb.append(spc+"factual="+f.factual) ;
         sb.append(spc) ;
         sb.append(spc+"subfactual="+f.subfactual) ;
         sb.append(spc) ;
         sb.append(nl) ;
      }
      sb.append("      =>") ; sb.append(nl) ;
      Disjunction d = r.consequent ;
      Iterator dit = d.iterator() ;
      while (dit.hasNext()) { 
         c = (Conjunction)dit.next() ;
         sb.append(spc+spc+spc+spc+spc+spc+"<conjunction>.factual="+c.factual+nl) ;
         for (int i = 0 ; i < c.size() ; i++) {
            sb.append(spc) ; sb.append(spc) ; sb.append(spc) ;
            sb.append(spc) ; sb.append(spc) ; sb.append(spc) ;
            Functor f = (Functor)c.elementAt(i) ; 
            sb.append(i+" key="+string(f.key)) ;
            sb.append(spc) ;
            sb.append(spc+"order="+string(f.order)) ;
            sb.append(spc) ;    
            sb.append(spc+"factual="+f.factual) ;
            sb.append(spc) ;
            sb.append(spc+"subfactual="+f.subfactual) ;
            sb.append(spc) ;
            sb.append(nl) ;
         }
            if (dit.hasNext()) { 
               sb.append(spc) ; sb.append(spc) ; sb.append(spc) ;
               sb.append(spc) ; sb.append(spc) ; sb.append(spc) ;
               sb.append("   |") ; sb.append(nl) ; 
             }
      }
      return sb ; 
   }

   // string shows predicate key
   private static String string(boolean[] k) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < k.length ; i++) {
         sb.append(k[i]) ; 
         if (i < k.length -1) sb.append(",") ; 
     }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   // string shows predicate order
   private static String string(int[] a) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < a.length ; i++) {
         sb.append(a[i]) ; 
         if (i < a.length -1) sb.append(",") ; 
      }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   // string shows PatternVector
   private static String string(Vector<Pattern> v) { 
      if (v == null) return " " ;
      StringBuffer sb = new StringBuffer() ;  
      sb.append("keys=<") ; 
      for(int i = 0 ; i < v.size() ; i++) {
         sb.append(string(((Pattern)v.elementAt(i)).key)) ;
         //sb.append("\n") ; 
      }
      sb.append(">") ; 
      return sb.toString() ; 
   }


}  // end CologAnalysis

