package sam.util ; 

/**
  *  Experiment to make a faster, smaller, reusable
  *  circular queue of nonneg int indices --
  *  no objects allowed.  15 FEB 2009
  *  Double capacity when overflow imminent.
  *  Seems about 10%+ faster, probably a lot less space.
  */
public class Q {

   private int[] data ;   // the data queue
   private int front, rear, size, capacity ;  

   /**
     *  Initialize Q, with initial capacity.
     *  Double capacity upon overflow.
     */
   public Q(int cap) { 
      this.capacity = cap ;  // initially
      this.data = new int[cap] ; 
      this.front = 0 ; 
      this.rear = 0 ; 
      this.size = 0 ; // empty
   }

   /**
     * @return the current size of this Q, 
     *  how many indices it holds.
     */
   public int size() { return this.size ; }

   /*** space saver */
   public void trimToSize() { 
      int[] ndata = new int[this.size] ; 
      for (int i = 0 ; i < this.size ; i++) {
         ndata[i] = this.data[i] ; 
      }
      this.data = ndata ; 
      this.capacity = this.size ;
   }


   /** 
     *  Clear this Q.
     */
   public void clear() { 
      this.front = 0 ;
      this.rear = 0 ;
      this.size = 0 ;
   }

   /** 
     *  Add to rear of Q.
     *  Expand Q 2x first if full.
     */
   public void add(int a) { 
      // expand Q if necessary
      if (size == capacity) { // expand first
         int[] d = new int[2*capacity] ; 
         // copy over data to d
         int k = 0 ; 
         if (front > 0) {
            for(int i = front ; i < capacity ; i++) {
               d[k]=data[i] ;
               k++ ;
            }
            for (int j = 0 ; j <=rear ; j++) {
               d[k]=data[j] ;
               k++ ;
             }
         }
         else { 
             for(int i = 0 ; i <capacity ; i++) {
                d[k]=data[i] ;
                k++ ;
             }
         }
         // now add
         d[capacity] = a ;
         data = d ; 
         front = 0 ; 
         rear = capacity ; 
         // now update capacity
         capacity = 2* capacity ;
      } // end case of add with expansion of circular Q
      else { 
         if(size == 0) {   // front=rear, a goes here
            data[rear] = a ;
            size++ ; 
         }
         else if (rear == capacity-1) {  // wrap around
            rear = 0 ; 
            data[rear] = a ;
            size++ ;
         }
         else {  // move right ->
            rear++ ;
            data[rear] = a ; 
            size++ ;
         }
      }
   }

   public boolean isEmpty() { 
      return this.size == 0 ;
   }

   /**
     *  Remove from front of Q.
     *  Q is intended to hold nonnegative leaf indices, so
     *  @return front of Q (int >=0) or -1 if empty 
     */
   public int remove() { 
      if (this.size == 0) return -1 ; 
      // otherwise
      size-- ; // new size
      int ret = data[front] ; 
      if (front != rear) {
         if (front == capacity-1) front = 0 ; // wrap around
         else front++ ;  // move right
      // In case front=rear, do nothing
      }
      return ret ; 
   }

   public int peek() { 
      return data[front] ; 
   } 

   // for testing 
   public void inspect() { 
      System.out.print(front+"{ ") ;
      for (int i=0;i<capacity;i++)
         System.out.print(data[i]+" ") ; 
      System.out.print(" }"+rear) ;
      System.out.println() ; 
   }      

   /** 
     *  Include other Q at rear of this one. 
     *  In effect, clears the included Q.
     */
   public void include(Q q) { 
      int s = q.size() ; // refer to before removing
      for(int i = 0 ; i < s ; i++) 
         this.add(q.remove()) ; 
   }

   /** 
     *  For circular usage. 
     *  Reset to initial state.
     */
   public void reset() { 
      this.front = 0 ; 
      this.rear = this.size -1 ; 
   }

   /**
     * Make a copy of this Q.
      */
   public Q clone() {
      Q nq = new Q(this.capacity) ; 
      for (int i = 0 ; i<this.capacity ; i++)
         nq.data[i] = this.data[i] ; 
      nq.front = this.front ; 
      nq.rear = this.rear ; 
      nq.size = this.size ; 
      return nq ;
   }

   public static void main(String[] args) { 
      Q q = new Q(10) ; 
      for (int n = 1 ; n < 11 ; n++)  
         q.add(n) ;
      Q qc = q.clone() ;
      q.clear() ;
      q.inspect() ; qc.inspect() ; 
/*
      Q q1 = new Q(10) ; 
      q1.add(1) ; 
      q1.add(2) ; 
      q1.add(3) ; 
      q1.inspect() ; 
      q1.trimToSize() ; 
      q1.inspect() ; 
      for (int t=0 ; t<14 ; t++) { 
         int k = q1.remove() ; 
         q1.add(k) ; 
         System.out.print(k+" ") ; 
         q1.inspect() ; 
      }
      q1.reset() ; 
      q1.inspect() ; 





      Q q = new Q(7) ;
      // time many add and remove
      long start =  System.currentTimeMillis() ; // milliseconds 
      int k = 0 ; 
      while (k < 100000) {
         for(int i = 0 ; i < 12 ; i++) 
            q.add(i) ; 
         for(int i = 0 ; i < 12 ; i++) 
            q.remove() ; 
         k++ ; 
      }
      System.out.println(System.currentTimeMillis()-start) ;
      for(int i = 0 ; i < 12 ; i++) 
            q.add(i) ; 
      q.inspect() ; 
*/
   }
   // VERY FAST about 14 time faster than Vector<Integer> used as queue ;-)
}

